package owl.img.analysis;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.geom.Point2D;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.EtchedBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.panel.CrosshairOverlay;
import org.jfree.chart.plot.Crosshair;

import arc.api.device.ArcDeviceJAPI;
import arc.api.image.ArcImageJAPI;
import owl.img.buffer.ImageBufferListener;
import owl.img.buffer.ImageUpdateEvent;
import owl.main.MainApp;
import owl.main.owltypes.OwlNumberSpinner;
import owl.main.setup.SetupEvent;
import owl.main.setup.SetupListener;
import owl.plot.HistogramPlot;
import owl.plot.PlotFactory;



public class HistogramPlotFrame extends AnalysisFrame implements SetupListener, ChangeListener, ImageBufferListener
{
	private static final long serialVersionUID = -4462784123633024973L;

	protected OwlNumberSpinner	m_rowStartSpinner;
	protected OwlNumberSpinner	m_rowEndSpinner;
	protected OwlNumberSpinner	m_colStartSpinner;
	protected OwlNumberSpinner	m_colEndSpinner;
	protected CrosshairOverlay	m_crosshairOverlay;
	protected HistogramPlot		m_histogramPlot;

	private   boolean			m_bChangesEnabled;
	private   Object			m_syncLock;


	public HistogramPlotFrame()
	{
		super( "Image Histogram" );

		m_syncLock			= new Object();
		m_bChangesEnabled	= true;

		m_histogramPlot = new HistogramPlot();
		m_histogramPlot.setXLabel( "ADU" );
		m_histogramPlot.setYLabel( "Pixel Count" );

		Crosshair crosshair = new Crosshair();

		m_crosshairOverlay = new CrosshairOverlay();
		m_crosshairOverlay.addDomainCrosshair( crosshair );
		m_crosshairOverlay.addRangeCrosshair( crosshair );

		ChartPanel chartPanel =
						PlotFactory.createHistogramChart( m_histogramPlot,
														  super.getTitle(),
														  Color.white,
														  true );
		chartPanel.addChartMouseListener( this );
		chartPanel.addOverlay( m_crosshairOverlay );

		super.setPlot( chartPanel );

		super.addComponent( createCtrlPanel(),
							super.SOUTH_CONTAINER_INDEX );

		super.setSize( new Dimension( 850, 550 ) );

		super.pack();

		loadPreferences();

		owl.main.setup.SetupFrame.addSetupListener( this );

		( new Thread( new ReadCtlrRunnable() ) ).start();

		ImageUpdateEvent.addImageBufferListener( this );
	}

	// +-----------------------------------------------------------------------------------+
	// | SetupListener method overrides                                                    |
	// +-----------------------------------------------------------------------------------+
	@Override
	public void setupChanged( SetupEvent event )
	{
		if ( event.bSetupOk )
		{
			setImageRowsValue( event.dRows );
			setImageColsValue( event.dCols );

			m_rowStartSpinner.setMaximum( event.dRows );
			m_rowEndSpinner.setMaximum( event.dRows );

			m_colStartSpinner.setMaximum( event.dCols );
			m_colEndSpinner.setMaximum( event.dCols );

			if ( ( ( Integer )m_rowStartSpinner.getValue() ).intValue() > event.dRows )
			{
				m_rowStartSpinner.setValue( Integer.valueOf( event.dRows ) );
			}

			if ( ( ( Integer )m_rowEndSpinner.getValue() ).intValue() > event.dRows ||
			   ( ( ( Integer )m_rowEndSpinner.getValue() ).intValue() == 0 ) )
			{
				m_rowEndSpinner.setValue( Integer.valueOf( event.dRows ) );
			}

			if ( ( ( Integer )m_colStartSpinner.getValue() ).intValue() > event.dCols )
			{
				m_colStartSpinner.setValue( Integer.valueOf( event.dCols ) );
			}

			if ( ( ( Integer )m_colEndSpinner.getValue() ).intValue() > event.dCols ||
			   ( ( ( Integer )m_colEndSpinner.getValue() ).intValue() == 0 ) )
			{
				m_colEndSpinner.setValue( Integer.valueOf( event.dCols ) );
			}
		}
	}

	// +-----------------------------------------------------------------------------------+
	// | ChangeListener method overrides                                                   |
	// +-----------------------------------------------------------------------------------+
	@Override
	public void stateChanged( ChangeEvent ce )
	{
		if ( m_bChangesEnabled )
		{
			if ( ( ce.getSource() instanceof javax.swing.JSpinner ) && this.isVisible() )
			{
				( new Thread( this ) ).start();
			}
		}
	}

	// +-----------------------------------------------------------------------------------+
	// | ImageBufferListener method overrides                                              |
	// +-----------------------------------------------------------------------------------+
	@Override
	public void bufferChanged( ImageUpdateEvent event )
	{
		boolean bAuto = false;

		enableChangeListeners( false );

		m_rowStartSpinner.setMaximum( ( int )event.getRows() );

		m_rowEndSpinner.setMaximum( ( int )event.getRows() );

		if ( ( ( Integer )m_rowEndSpinner.getValue() ).intValue() > event.getRows() )
		{
			m_rowEndSpinner.setValue( Integer.valueOf( ( int )event.getRows() ) );

			bAuto = true;
		}

		m_colStartSpinner.setMaximum( ( int )event.getCols() );

		m_colEndSpinner.setMaximum( ( int )event.getCols() );

		if ( ( ( Integer )m_colEndSpinner.getValue() ).intValue() > event.getCols() )
		{
			m_colEndSpinner.setValue( Integer.valueOf( ( int )event.getCols() ) );

			bAuto = true;
		}

		m_rowsTxtfld.setValue( event.getRows() );
		m_colsTxtfld.setValue( event.getCols() );

		enableChangeListeners( true );

		if ( this.isVisible() )
		{
			( new Thread( this ) ).start();
		}

		if ( bAuto )
		{
			setAutoRange();
		}
	}

	// +-----------------------------------------------------------------------------------+
	// | AnalysisFrame method overrides                                                    |
	// +-----------------------------------------------------------------------------------+
	@Override
	public void readCtlr()
	{
		enableChangeListeners( false );

		try
		{
			if ( ArcDeviceJAPI.IsOpen() && ArcDeviceJAPI.IsControllerConnected() )
			{
				MainApp.infoStart( "Reading image dimensions from controller" );

				int dRowEnd = ArcDeviceJAPI.GetImageRows();
				int dColEnd = ArcDeviceJAPI.GetImageCols();

				setImageRowsValue( dRowEnd );
				setImageColsValue( dColEnd );

				m_rowStartSpinner.setMaximum( dRowEnd );
				m_rowStartSpinner.setValue( Integer.valueOf( 0 ) );

				m_rowEndSpinner.setMaximum( dRowEnd );
				m_rowEndSpinner.setValue( dRowEnd );

				m_colStartSpinner.setMaximum( dColEnd );
				m_colStartSpinner.setValue( Integer.valueOf( 0 ) );

				m_colEndSpinner.setMaximum( dColEnd );
				m_colEndSpinner.setValue( dColEnd );

				MainApp.infoEnd();
			}
			else
			{
				MainApp.error( "( ReadCtlrRunnable ): " +
							   "No device or controller connected!" );
			}
		}
		catch ( Throwable e )
		{
			MainApp.error( e );
		}

		super.pack();

		enableChangeListeners( true );
	}

	@Override
	public void readDS9()
	{
		try
		{
			MainApp.info( "Reading image dimensions from DS9 ...." );
			MainApp.warn( "Requires POINT shape to be manually selected within DS9!" );

			long[] lCoords = getDS9Box();
			
			if ( m_bStop ) { m_runButton.setBlack( RUN_ACTION ); return; }
			
			long lRowStart = ( lCoords[ BOX_CENTER_ROW ] - ( lCoords[ BOX_ROWS ] / 2 ) );
			long lRowEnd   = ( lCoords[ BOX_CENTER_ROW ] + ( lCoords[ BOX_ROWS ] / 2 ) );
			long lColStart = ( lCoords[ BOX_CENTER_COL ] - ( lCoords[ BOX_COLS ] / 2 ) );
			long lColEnd   = ( lCoords[ BOX_CENTER_COL ] + ( lCoords[ BOX_COLS ] / 2 ) );

			setImageRowsValue( MainApp.getMainFrame().getCameraPanel().getRowLabelValue() );
			setImageColsValue( MainApp.getMainFrame().getCameraPanel().getColLabelValue() );

			m_rowStartSpinner.setValue( ( new Long( lRowStart ) ).intValue() );
			m_rowEndSpinner.setValue( ( new Long( lRowEnd ) ).intValue() );

			m_colStartSpinner.setValue( ( new Long( lColStart ) ).intValue() );
			m_colEndSpinner.setValue( ( new Long( lColEnd ) ).intValue() );
		}
		catch ( Throwable e )
		{
			MainApp.error( e );
		}
	}

	@Override
	public void setAutoRange()
	{
		m_histogramPlot.getDomainAxis().setAutoRange( true );
		m_histogramPlot.getRangeAxis().setAutoRange( true );
	}

	@Override
	public void run()
	{
		Thread.currentThread().setName( "Owl - Histogram" );

		m_runButton.setColored( ABORT_ACTION );

		try
		{
			synchronized( m_syncLock )
			{
				if ( m_bStop )
				{
					throw new Exception( "Plot update aborted! You should not see this!" );
				}

				Point2D.Float[] points =
								ArcImageJAPI.Histogram( ArcDeviceJAPI.GetCommonBufferVA(),
														m_rowStartSpinner.getInt(),
														m_rowEndSpinner.getInt(),
														m_colStartSpinner.getInt(),
														m_colEndSpinner.getInt(),
													    super.getImageRowsValue(),
													    super.getImageColsValue() );

				m_histogramPlot.setPlot( points, "Histogram" );
	
				if ( m_bStop )
				{
					throw new Exception( "Plot update aborted! You should not see this!" );
				}
	
				super.getStats( m_rowStartSpinner.getInt(),
								m_rowEndSpinner.getInt(),
								m_colStartSpinner.getInt(),
								m_colEndSpinner.getInt(),
								super.getImageRowsValue(),
								super.getImageColsValue() );
			}	// end synchronization code
		}
		catch ( NumberFormatException nfe )
		{
			MainApp.error( "( ImageHistrogramPlot ): " + nfe.getMessage() );
		}
		catch ( Throwable e )
		{
			if ( !m_bStop )
			{
				MainApp.error( "( ImageHistrogramPlot ): " + e.getMessage() );
			}
		}
		finally
		{
			m_runButton.setBlack( RUN_ACTION );
		}

		super.pack();
	}

	// +-----------------------------------------------------------------------------------+
	// | createCtrlPanel                                                                   |
	// +-----------------------------------------------------------------------------------+
	// | Creates the plot control ( input ) panel components.                              |
	// +-----------------------------------------------------------------------------------+
	private JPanel createCtrlPanel()
	{
		JPanel rPanel = new JPanel();
		rPanel.setBorder( BorderFactory.createEtchedBorder( EtchedBorder.RAISED ) );

		m_rowStartSpinner = new OwlNumberSpinner( 6 );
		m_rowStartSpinner.addChangeListener( this );

		m_rowEndSpinner = new OwlNumberSpinner( 6 );
		m_rowEndSpinner.addChangeListener( this );

		rPanel.add( new JLabel( "<html><b>ROW</b></html>" ) );
		rPanel.add( javax.swing.Box.createHorizontalStrut( 5 ) );
		rPanel.add( new JLabel( "start: " ) );
		rPanel.add( m_rowStartSpinner );
		rPanel.add( javax.swing.Box.createHorizontalStrut( 5 ) );
		rPanel.add( new JLabel( "end: " ) );
		rPanel.add( m_rowEndSpinner );

		JPanel cPanel = new JPanel();
		cPanel.setBorder( BorderFactory.createEtchedBorder( EtchedBorder.RAISED ) );

		m_colStartSpinner = new OwlNumberSpinner( 6 );
		m_colStartSpinner.addChangeListener( this );

		m_colEndSpinner = new OwlNumberSpinner( 6 );
		m_colEndSpinner.addChangeListener( this );

		cPanel.add( new JLabel( "<html><b>COL</b></html>" ) );
		cPanel.add( javax.swing.Box.createHorizontalStrut( 5 ) );
		cPanel.add( new JLabel( "start: " ) );
		cPanel.add( m_colStartSpinner );
		cPanel.add( javax.swing.Box.createHorizontalStrut( 5 ) );
		cPanel.add( new JLabel( "end: " ) );
		cPanel.add( m_colEndSpinner );

		JPanel panel = new JPanel();
		panel.setBorder( BorderFactory.createLoweredBevelBorder() );
		panel.setLayout( super.gbl );

		addComponent( panel, rPanel,              0, 0, 0, 0, 0, 0, 1, 1 );
		addComponent( panel, cPanel,              0, 0, 0, 0, 0, 1, 1, 1 );
		addComponent( panel, getCoordsSrcPanel(), 0, 0, 0, 0, 0, 2, 1, 1 );
		addComponent( panel, createStatsPanel(),  0, 0, 5, 0, 1, 0, 1, 3 );

		return panel;
	}

	// +-----------------------------------------------------------------------------------+
	// | enableChangeListeners                                                             |
	// +-----------------------------------------------------------------------------------+
	// | Enables/Disables the change listeners for the JSpinner components in this window. |
	// | Used to disable spinner change listeners at startup.                              |
	// +-----------------------------------------------------------------------------------+
	private void enableChangeListeners( boolean bEnable )
	{
		m_bChangesEnabled = bEnable;
	}
}
