package owl.img.buffer;

import java.util.EventObject;
import java.util.Vector;


// +------------------------------------------------------------------------------------+
// | ImageUpdateEvent Class                                                             |
// +------------------------------------------------------------------------------------+
// | Sent to any ImageBufferListener to inform it that the image buffer data has        |
// | updated.                                                                           |
// +------------------------------------------------------------------------------------+
public class ImageUpdateEvent extends EventObject
{
	private static final long serialVersionUID = 7479811144637595391L;

	private String	m_sFits;
	private long	m_lRows;
	private long	m_lCols;


	// +--------------------------------------------------------------------------------+
	// | Constructor                                                                    |
	// +--------------------------------------------------------------------------------+
	// | @param source - The source of the event.                                       |
	// | @param lRows  - The new image row size.                                        |
	// | @param lCols  - The new image col size.                                        |
	// +--------------------------------------------------------------------------------+
	public ImageUpdateEvent( Object oSource, long lRows, long lCols )
	{
		super( oSource );

		m_lRows	= lRows;
		m_lCols	= lCols;
		m_sFits = null;
	}

	// +--------------------------------------------------------------------------------+
	// | Constructor                                                                    |
	// +--------------------------------------------------------------------------------+
	// | @param source - The source of the event.                                       |
	// | @param lRows  - The new image row size.                                        |
	// | @param lCols  - The new image col size.                                        |
	// | @param sFits  - The full path and name of the loaded FITS file.                |
	// +--------------------------------------------------------------------------------+
	public ImageUpdateEvent( Object oSource, long lRows, long lCols, String sFits )
	{
		super( oSource );

		m_lRows	= lRows;
		m_lCols	= lCols;
		m_sFits = sFits;
	}

	// +--------------------------------------------------------------------------------+
	// | getRows                                                                        |
	// +--------------------------------------------------------------------------------+
	// | Returns the new image row dimension of the data within the buffer.             |
	// +--------------------------------------------------------------------------------+
	public long getRows()
	{
		return m_lRows;
	}

	// +--------------------------------------------------------------------------------+
	// | getCols                                                                        |
	// +--------------------------------------------------------------------------------+
	// | Returns the new image column dimension of the data within the buffer.          |
	// +--------------------------------------------------------------------------------+
	public long getCols()
	{
		return m_lCols;
	}

	// +--------------------------------------------------------------------------------+
	// | getFits                                                                        |
	// +--------------------------------------------------------------------------------+
	// | Returns the full path and name of the loaded FITS file.                        |
	// +--------------------------------------------------------------------------------+
	public String getFits()
	{
		return m_sFits;
	}

	// +--------------------------------------------------------------------------------+
	// | Static Class Members/Methods                                                   |
	// +--------------------------------------------------------------------------------+
	private static Vector<ImageBufferListener> sm_vListeners = new Vector<ImageBufferListener>();

	// +--------------------------------------------------------------------------------+
	// | Static - callImageBufferListeners                                              |
	// +--------------------------------------------------------------------------------+
	// | Adds the image buffer listeners to the call list.                              |
	// |                                                                                |
	// | @param l - The ImageBufferListener to add.                                     |
	// +--------------------------------------------------------------------------------+
	public static void addImageBufferListener( ImageBufferListener l )
	{
		if ( !sm_vListeners.contains( l ) )
		{
			sm_vListeners.add( l );
		}
	}

	// +--------------------------------------------------------------------------------+
	// | Static - callImageBufferListeners                                              |
	// +--------------------------------------------------------------------------------+
	// | Calls all the image buffer listeners with the new image dimensions.            |
	// |                                                                                |
	// | @param event - The ImageUpdateEvent to send to all listeners.                  |
	// +--------------------------------------------------------------------------------+
	public static void callImageBufferListeners( ImageUpdateEvent event )
	{
		for ( ImageBufferListener l : sm_vListeners )
		{
			l.bufferChanged( event );
		}
	}
}
