package owl.main.exposure;

import java.io.File;

import arc.api.deinterlace.ArcDeinterlaceJAPI;
import arc.api.device.ArcReplyException;
import arc.api.device.ArcDeviceJAPI;
import arc.api.fits.ArcFitsFileJAPI;

import owl.gui.utils.OwlUtilities;
import owl.main.MainApp;


public class ExposeRunnableCR extends ExposeRunnable
{
	private final int FPB	=	0x465042;
	private final int SNF	=	0x534E46;

	private int  m_numOfFrames;
	private long m_hFits;

	public ExposeRunnableCR( int numOfFrames, ExposeInfo expInfo, ExposeActionListener expActListener )
	{
		super( expInfo, expActListener );
		m_numOfFrames = numOfFrames;
	}

	@Override
	public void run()
	{
		long lImageSize         = 0;
		long lBoundedImageSize  = 0;
		long lFramesPerBuffer   = 0;
		long lPCIFrameCount     = 0;
		long lLastPCIFrameCount = 0;
		long lFPBCount          = 0;
		long lFitsFrameCount	= 0;
		long lReadoutStartTime	= 0;

		m_hFits = 0;

		Thread.currentThread().setName( "Owl - ExposeRunnableCR" );

		try
		{
			if ( m_exposeActionListener != null )
			{
				m_exposeActionListener.setExposeAction( ExposeActionListener.ABORT );
			}

			// Pre-fill image buffer if requested
			// --------------------------------------------------------------
			if ( m_exposeInfo.isFillBuffer() )
			{
				MainApp.infoStart( "Filling image buffer" );
				ArcDeviceJAPI.FillCommonBuffer( m_exposeInfo.getBufferFill() );
				MainApp.infoEnd();
			}

			MainApp.infoStart( "Initializing continuous readout" );

			m_exposeInfo.randomPixelCountColor( java.awt.Color.black );
			m_exposeInfo.setPixelRange( 0, m_numOfFrames );
			m_exposeInfo.setFrameCount( 0 );

			if ( m_exposeInfo.isDisplay() )
			{
				MainApp.warn( "Image display not supported during continuous readout!" );
			}

			// Check for valid frame count
			if ( m_numOfFrames <= 0 )
			{
				throw new Exception( "Number of frames must be > 0" );
			}
	
			// Pre-get deinterlacing algorithm and channel count ( optional )
			// --------------------------------------------------------------
			int algorithm	= m_exposeInfo.getDeinterlaceAlgorithm();
			int arg			= -1;

			if ( algorithm == ArcDeinterlaceJAPI.DEINTERLACE_HAWAII_RG )
			{
				arg = ArcDeviceJAPI.Cmd2( ArcDeviceJAPI.TIM_ID, ArcDeviceJAPI.RNC );
			}

			// Get the image size
			// --------------------------------------------------------------
			int[] dImageSize = ArcDeviceJAPI.GetImageSize();
			int dRows = dImageSize[ 0 ];
			int dCols = dImageSize[ 1 ];
			lImageSize = dRows * dCols * 2;
	
			if ( !ArcDeviceJAPI.ToString().contains( "PCIe" ) && ( lImageSize & 0x3FF ) != 0 )
			{
				lBoundedImageSize = lImageSize - ( lImageSize & 0x3FF ) + 1024;
			}
			else
			{
				lBoundedImageSize = lImageSize;
			}
	
			if ( m_bStop ) { throw new Exception( STOP_MSG ); }
	
			if ( lBoundedImageSize <= 0 )
			{
				throw new Exception( "Calculated an image size of zero!" );
			}

			lFramesPerBuffer = ( long )Math.floor(
								( ArcDeviceJAPI.GetCommonBufferSize() / lBoundedImageSize ) );
	
			if ( m_bStop ) { throw new Exception( STOP_MSG ); }

			// Determine the max frames per FITS file
			// -------------------------------------------------------------
			long lMaxFitsFrames =
							( long )Math.floor( Math.pow( 2, 31 ) / lBoundedImageSize ) - 1;

			// Get FITS filename and check for valid extension
			// -------------------------------------------------------------
			if ( !m_exposeInfo.getFileExtension().equals( ExposePanel.FITS_FILEEXT_1 ) )
			{
				throw new Exception( "File type MUST be FITS! Currently set to: " +
									  m_exposeInfo.getFileExtension() );
			}

			if ( m_exposeInfo.isIncrementFilename() )
			{
				m_exposeInfo.incrementFilename();
			}

			String fitsFilename = m_exposeInfo.getFilename();

			MainApp.infoEnd();

			// Create the FITS data cube
			// -------------------------------------------------------------
			MainApp.infoStart( "Creating FITS data cube" );
			m_hFits = ArcFitsFileJAPI.Create3D( fitsFilename,
											    dRows,
											    dCols,
											    ArcFitsFileJAPI.BPP16 );
			MainApp.infoEnd();

			// Set the frames-per-buffer
			MainApp.infoStart( "Setting frames-per-buffer to " + lFramesPerBuffer );
			ArcDeviceJAPI.Cmd( ArcDeviceJAPI.TIM_ID, FPB, ( int )lFramesPerBuffer, ArcDeviceJAPI.DON );
			MainApp.infoEnd();

			if ( m_bStop ) { throw new Exception( STOP_MSG ); }

			// Set the number of frames-to-take
			MainApp.infoStart( "Setting number-of-frames to " + m_numOfFrames );
			ArcDeviceJAPI.Cmd( ArcDeviceJAPI.TIM_ID, SNF, m_numOfFrames, ArcDeviceJAPI.DON );
			MainApp.infoEnd();

			if ( m_bStop ) { throw new Exception( STOP_MSG ); }

			// Set the exposure time
			// --------------------------------------------------------------
			int dExpTime = ( int )( m_exposeInfo.getExposeTime() * 1000.0 );

			MainApp.infoStart( "Setting exposure time" );
			ArcDeviceJAPI.Cmd( ArcDeviceJAPI.TIM_ID, ArcDeviceJAPI.SET, dExpTime, ArcDeviceJAPI.DON );
			MainApp.infoEnd();

			if ( m_bStop ) { throw new Exception( STOP_MSG ); }

			// Start the exposure
			// --------------------------------------------------------------
			MainApp.infoStart( "Starting exposure " );
			ArcDeviceJAPI.Cmd( ArcDeviceJAPI.TIM_ID, ArcDeviceJAPI.SEX, ArcDeviceJAPI.DON );
			MainApp.infoEnd();

			if ( m_bStop ) { throw new Exception( STOP_MSG ); }

			lReadoutStartTime = System.currentTimeMillis();

			// Read the images
			while ( lPCIFrameCount < m_numOfFrames )
			{
				if ( m_bStop ) { throw new Exception( STOP_MSG ); }

				lPCIFrameCount = ArcDeviceJAPI.GetFrameCount();

				if ( m_bStop ) { throw new Exception( STOP_MSG ); }

				if ( lFitsFrameCount > lMaxFitsFrames )
				{
					ArcFitsFileJAPI.Close( m_hFits );

					lFitsFrameCount = 0;

					fitsFilename = OwlUtilities.filenameIncr( new File( fitsFilename ) );

					m_hFits = ArcFitsFileJAPI.Create3D( fitsFilename,
													    dRows,
													    dCols,
													    ArcFitsFileJAPI.BPP16 );
				}

				if ( m_bStop ) { throw new Exception( STOP_MSG ); }

				if ( lPCIFrameCount > lLastPCIFrameCount )
				{
					if ( lFPBCount >= lFramesPerBuffer )
					{
						lFPBCount = 0;
					}

					// Deinterlace image
					// --------------------------------------------------------------
					ArcDeinterlaceJAPI.RunAlg( ArcDeviceJAPI.GetCommonBufferVA(),
											   dRows,
											   dCols,
											   ( int )( lFPBCount * ( lBoundedImageSize / 2 ) ),
											   algorithm,
											   arg );

					if ( m_bStop ) { throw new Exception( STOP_MSG ); }

					// Write FITS file
					// --------------------------------------------------------------
					ArcFitsFileJAPI.Write3D( m_hFits,
											 ArcDeviceJAPI.GetCommonBufferVA(),
											 ( int )( lFPBCount * lBoundedImageSize ) );

					m_exposeInfo.setFrameCount( lPCIFrameCount );

					m_exposeInfo.setReadoutTime( ( System.currentTimeMillis() -
						      					  lReadoutStartTime ) / 1000.0 );

					lLastPCIFrameCount = lPCIFrameCount;
					lFitsFrameCount++;
					lFPBCount++;
				}
			}

			// Deinterlace image
			// --------------------------------------------------------------
//			ArcFitsFileJAPI.Close( m_hFits );
//
//			m_hFits = ArcFitsFileJAPI.Open( fitsFilename );
//
//			long m_hOutFits = ArcFitsFileJAPI.Create3D( fitsFilename.replace( ".fit", "OUT.fit" ), dRows, dCols, ArcFitsFileJAPI.BPP16 );
//			System.out.println( "FILE -> " + fitsFilename + " NOF -> " + ArcFitsFileJAPI.GetNumberOfFrames( m_hFits ) );
//
//			for ( int i=0; i<ArcFitsFileJAPI.GetNumberOfFrames( m_hFits ); i++ )
//			{
//				long lVA = ArcFitsFileJAPI.Read3D( m_hFits, i );
//
//				System.out.println( "FRAME -> " + i + " VA -> " + Long.toHexString( lVA ) );
//
//				ArcDeinterlaceJAPI.RunAlg( lVA,
//										   dRows,
//										   dCols,
//										   0,
//										   algorithm,
//										   arg );
//
//				ArcFitsFileJAPI.Write3D( m_hOutFits, i );
//			}

//			System.out.println( "lPCIFrameCount -> " + lPCIFrameCount );
//			System.out.println( "lLastPCIFrameCount -> " + lLastPCIFrameCount );
//			System.out.println( "lFitsFrameCount -> " + lFitsFrameCount );
		}
		catch ( ArcReplyException re )
		{
			MainApp.infoFail();
			MainApp.error( re.toString() );
		}
		catch ( Throwable e )
		{
			MainApp.infoFail();

			if ( e.getMessage().equals( STOP_MSG ) )
			{
				MainApp.warn( e.getMessage() );
			}
			else
			{
				MainApp.error( e );
			}
		}
		finally
		{
			cleanup();

			if ( m_exposeActionListener != null )
			{
				m_exposeActionListener.setExposeAction( ExposeActionListener.EXPOSE );
			}
		}
	}

	// +--------------------------------------------------------------------+
	// |  cleanup()                                                         |
	// +--------------------------------------------------------------------+
	// |  Performs any necessary cleanup, such as setting the camera back   |
	// |  to single readout mode.                                           |
	// +--------------------------------------------------------------------+
	protected void cleanup()
	{
		try
		{
			ArcFitsFileJAPI.Close( m_hFits );

			ArcDeviceJAPI.Cmd( ArcDeviceJAPI.TIM_ID, SNF, 1, ArcDeviceJAPI.DON );
		}
		catch ( ArcReplyException re ) { MainApp.error( re.toString() ); }
		catch ( Throwable e ) {	MainApp.error( e ); }
	}
}
