package owl.main.scripting;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;

import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.WindowConstants;

import owl.gui.utils.OwlUtilities;
import owl.main.MainApp;
import owl.main.owltypes.OwlFrame;

import bsh.Interpreter;
import bsh.NameSpace;
import bsh.util.JConsole;


public class BshConsoleFrame extends OwlFrame implements ActionListener
{
	private static final long serialVersionUID = 3209543809866802920L;

	private Interpreter	m_interpreter;
	private Thread		m_thread;


	public BshConsoleFrame()
	{
		super( "Owl Script Console", true );

		setDefaultCloseOperation( WindowConstants.DISPOSE_ON_CLOSE );
		setIconImage( Toolkit.getDefaultToolkit().getImage( MainApp.getBitmapPath() + "TinyBean.gif" ) );
		getContentPane().setLayout( new BorderLayout() );

		super.addComponent( createToolbar(), super.TOOLBAR_INDEX );
		super.addComponent( createConsolePanel(), super.CENTER_CONTAINER_INDEX );

		pack();

		OwlUtilities.centerFrame( this );
	}

	//--------------------------------------------------------------------------
    //   Public Methods:
    //--------------------------------------------------------------------------
	@Override
	public void actionPerformed( ActionEvent e )
	{
		super.actionPerformed( e );
	}

	//--------------------------------------------------------------------------
    //   Private Methods:
    //--------------------------------------------------------------------------
	private JToolBar createToolbar()
	{
		JToolBar toolbar = new JToolBar();

		toolbar.setFloatable( false );
		super.appendToolbar( toolbar );

		return toolbar;
	}

	private JPanel createConsolePanel()
	{
		JPanel panel = new JPanel();
		panel.setLayout( new BorderLayout() );

		JConsole bshConsole = new JConsole();
		bshConsole.setPreferredSize( new Dimension( 800, 400 ) );

		panel.add( bshConsole, BorderLayout.CENTER );

		try
		{
			m_interpreter = new Interpreter( bshConsole );
			m_interpreter.setNameSpace( getStdNameSpace( "Console" ) );
		}
		catch ( Exception e )
		{
			MainApp.error( e );
		}

		m_thread = new Thread( m_interpreter );
		m_thread.start();

		return panel;
	}


	// +---------------------------------------------------------------------------+
	// |  getStdNameSpace                                                          |
	// +---------------------------------------------------------------------------+
	// |  Returns a namespace that has been initialized with a pre-defined set of  |
	// |  classes, packages and variable names.                                    |
	// |                                                                           |
	// |  @param sName - The name of the namespace.                                |
	// +---------------------------------------------------------------------------+
	public bsh.NameSpace getStdNameSpace( String sName )
	{
		NameSpace nameSpace = new NameSpace( m_interpreter.getNameSpace(),
											 "NameSpace_" + sName );

		nameSpace.getClassManager().setClassLoader(
											ClassLoader.getSystemClassLoader() );

		try
		{
			nameSpace.getClassManager().addClassPath(
					( new java.io.File( MainApp.getAppPath() ) ).toURI().toURL() );

			nameSpace.getClassManager().addClassPath(
					( new java.io.File( MainApp.getScriptsPath() ) ).toURI().toURL() );

			setStdInit( nameSpace );
		}
		catch ( Throwable e )
		{
			MainApp.error( e.getMessage() );
		}

		return nameSpace;
	}


	// +---------------------------------------------------------------------------+
	// |  setStdInit                                                               |
	// +---------------------------------------------------------------------------+
	// |  Sets the specified namespace with a pre-defined standard set of classes, |
	// |  packages and variable names.                                             |
	// |                                                                           |
	// |  @param nameSpace - The namespace to be initialized.                      |
	// +---------------------------------------------------------------------------+
	public void setStdInit( NameSpace nameSpace )
	{
		try
		{
			//  Add the scripts folder path along with all sub-folder names
			// +-----------------------------------------------------------+
			File fScriptPath = new File( MainApp.getScriptsPath() );

			nameSpace.importCommands( fScriptPath.getName() );

			File[] fScriptPathList = fScriptPath.listFiles();

			for ( int i=0; i<fScriptPathList.length; i++ )
			{
				if ( fScriptPathList[ i ].isDirectory() )
				{
					nameSpace.importCommands( fScriptPathList[ i ].getName() );
				}
			}


			//  Add variables
			// +-----------------------------------------------------------+
			nameSpace.setVariable( "logger",     MainApp.getLogger(),     false );
			nameSpace.setVariable( "interp",     m_interpreter,           false );
			nameSpace.setVariable( "bitmapPath", MainApp.getBitmapPath(), false );
			nameSpace.setVariable( "xmlPath",    MainApp.getXMLPath(),    false );

			nameSpace.setVariable( "ns", nameSpace, false );

			//  Add default imports ( packages and classes )
			// +-----------------------------------------------------------+
			nameSpace.loadDefaultImports();


			//  Add packages
			// +-----------------------------------------------------------+
			nameSpace.importPackage( "java.awt.geom" );
			nameSpace.importPackage( "owl.plot" );


			//  Add classes
			// +-----------------------------------------------------------+
			nameSpace.importClass( "import arc.api.deinterlace.ArcDeinterlaceJAPI;" );
			nameSpace.importClass( "import arc.api.device.ArcDeviceJAPI;" );
			nameSpace.importClass( "import arc.api.device.ReplyException;" );
			nameSpace.importClass( "import arc.api.fits.ArcFitsFileJAPI;" );
			nameSpace.importClass( "import arc.api.tiff.ArcTiffFileJAPI;" );
			nameSpace.importClass( "import arc.api.image.ArcImageJAPI;" );

			nameSpace.importClass( "import owl.main.fits.FitsHeaderFrame;" );
			nameSpace.importClass( "import owl.main.MainApp;" );
		}
		catch ( Throwable e )
		{
			MainApp.error( e.getMessage() );
		}
	}
}
