package owl.main.scripting;


import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Random;

import owl.main.MainApp;

import bsh.Interpreter;
import bsh.NameSpace;



// +-------------------------------------------------------------------------------+
// |  MainInterpreter Class                                                        |
// +-------------------------------------------------------------------------------+
// |  Provides a single instance of bsh.Interpreter for running BeanShell scripts  |
// |  within Owl.  Each script ( or script statement ) is executed within its own  |
// |  namespace ( to prevent variable name conflicts ).                            |
// +-------------------------------------------------------------------------------+
public class MainInterpreter
{
    //--------------------------------------------------------------------------
    //   Private Members:
    //--------------------------------------------------------------------------
    private static Interpreter		m_interpreter		= null;
    private static Random			m_random			= null;
    private static ArrayList<URL>	m_externClassPaths	= null;

 	static
	{
		m_interpreter		= new Interpreter();
		m_random			= new Random( System.currentTimeMillis() );
		m_externClassPaths	= new ArrayList<URL>();

		m_interpreter.setClassLoader( ClassLoader.getSystemClassLoader() );
	}

	// +---------------------------------------------------------------------------+
	// |  get                                                                      |
	// +---------------------------------------------------------------------------+
	// |  Returns a reference to the bsh.Interpreter.                              |
	// +---------------------------------------------------------------------------+
	public static bsh.Interpreter get()
	{
		return m_interpreter;
	}

	// +---------------------------------------------------------------------------+
	// |  source                                                                   |
	// +---------------------------------------------------------------------------+
	// |  Executes the specified script file.                                      |
	// |                                                                           |
	// |  @param sStatement - The statement(s) to execute.                         |
	// +---------------------------------------------------------------------------+
	public static Object source( String sFileName ) throws Throwable
	{
		String sName = createNameSpaceName( ( new File( sFileName ) ).getName() );

		bsh.NameSpace nameSpace = getStdNameSpace( sName );

//		Object[][] nsInfo = MainInterpreter.getNameSpaceInfo( nameSpace );
//		System.out.println( "NAMESPACE -> " + nameSpace.getName() );
//		System.out.println( "===================================================================" );
//		for ( Object[] nsArr : nsInfo )
//		{
//			if ( nsArr != null )
//			{
//				System.out.println( ( String )nsArr[ 0 ] + " -> " + nsArr[ 1 ] );
//			}
//		}
//		System.out.println( " " );

		return m_interpreter.source( sFileName, nameSpace );
	}

	// +---------------------------------------------------------------------------+
	// |  source                                                                   |
	// +---------------------------------------------------------------------------+
	// |  Executes the specified script file with the specified variable names and |
	// |  objects injected into the namespace.                                     |
	// |                                                                           |
	// |  @param sFileName  - The script file to execute.                          |
	// |  @param varNames   - The variable names to be included in the namespace.  |
	// |  @param varObjects - The variable objects to be included in the namespace.|
	// +---------------------------------------------------------------------------+
	public static Object source( String sFileName, String[] varNames,
								 Object[] varObjects ) throws Throwable
	{
		String sName = createNameSpaceName( ( new File( sFileName ) ).getName() );

		bsh.NameSpace nameSpace = getStdNameSpace( sName, varNames, varObjects );

//		Object[][] nsInfo = MainInterpreter.getNameSpaceInfo( nameSpace );
//		System.out.println( "NAMESPACE -> " + nameSpace.getName() );
//		System.out.println( "===================================================================" );
//		for ( Object[] nsArr : nsInfo )
//		{
//			if ( nsArr != null )
//			{
//				System.out.println( ( String )nsArr[ 0 ] + " -> " + nsArr[ 1 ] );
//			}
//		}
//		System.out.println( " " );

		return m_interpreter.source( sFileName, nameSpace );
	}

	// +---------------------------------------------------------------------------+
	// |  eval                                                                     |
	// +---------------------------------------------------------------------------+
	// |  Executes the specified script statement(s).                              |
	// |                                                                           |
	// |  @param sStatement - The statement(s) to execute.                         |
	// +---------------------------------------------------------------------------+
	public static Object eval( String sStatement ) throws Throwable
	{
		bsh.NameSpace nameSpace = getStdNameSpace( createNameSpaceName() );

		return m_interpreter.eval( sStatement, nameSpace );
	}

	// +---------------------------------------------------------------------------+
	// |  eval                                                                     |
	// +---------------------------------------------------------------------------+
	// |  Executes the specified script statement(s) with the specified variable   |
	// |  names and objects injected into the namespace.                           |
	// |                                                                           |
	// |  @param sStatement - The statement(s) to execute.                         |
	// |  @param varNames   - The variable names to be included in the namespace.  |
	// |  @param varObjects - The variable objects to be included in the namespace.|
	// +---------------------------------------------------------------------------+
	public static Object eval( String sStatement, String[] varNames,
							   Object[] varObjects ) throws Throwable
	{
		bsh.NameSpace nameSpace = getStdNameSpace( createNameSpaceName(),
												   varNames,
												   varObjects );

		return m_interpreter.eval( sStatement, nameSpace );
	}

	// +---------------------------------------------------------------------------+
	// |  containsVariable                                                         |
	// +---------------------------------------------------------------------------+
	// |  Searches the specified namespace for the specified variable name.        |
	// |  Returns 'true' if the name exists; 'false' otherwise.                    |
	// |                                                                           |
	// |  @param nameSpace - The namespace to search.                              |
	// |  @param varName   - The variable to search for.                           |
	// +---------------------------------------------------------------------------+
	public static boolean containsVariable( NameSpace nameSpace, String varName )
	{
		boolean bFound = false;

		String[] varNameArr = nameSpace.getVariableNames();

		for ( int i=0; i<varNameArr.length; i++ )
		{
			if ( varNameArr[ i ].equals( varName ) )
			{
				bFound = true;

				break;
			}
		}

		return bFound;
	}

	// +---------------------------------------------------------------------------+
	// |  addExternalClassPath                                                     |
	// +---------------------------------------------------------------------------+
	// |  Adds an external classpath to all namespaces.                            |
	// |                                                                           |
	// |  @param classPath - URL of the classpath to add.                          |
	// +---------------------------------------------------------------------------+
	public static void addExternalClassPath( URL classPath )
	{
		if ( !m_externClassPaths.contains( classPath ) )
		{
			m_externClassPaths.add( classPath );
		}
	}

	// +---------------------------------------------------------------------------+
	// |  removeExternalClassPath                                                  |
	// +---------------------------------------------------------------------------+
	// |  Removes an external classpath from all namespaces.                       |
	// |                                                                           |
	// |  @param classPath - URL of the classpath to add.                          |
	// +---------------------------------------------------------------------------+
	public static void removeExternalClassPath( URL classPath )
	{
		if ( m_externClassPaths.contains( classPath ) )
		{
			m_externClassPaths.remove( classPath );
		}
	}

	// +---------------------------------------------------------------------------+
	// |  getStdNameSpace                                                          |
	// +---------------------------------------------------------------------------+
	// |  Returns a namespace that has been initialized with a pre-defined set of  |
	// |  classes, packages and variable names as well as with the specified list  |
	// |  of variable names and objects.  Each name must have a corresponding      |
	// |  object value ( i.e. varNames.length must equal varObjects.length ).      |
	// |                                                                           |
	// |  @param sName      - The name of the namespace.                           |
	// |  @param varNames   - The list of variable names.                          |
	// |  @param varObjects - The list of variable objects.                        |
	// +---------------------------------------------------------------------------+
	public static bsh.NameSpace
	getStdNameSpace( String sName, String[] varNames, Object[] varObjects )
	{
		NameSpace nameSpace = getStdNameSpace( sName );

		try
		{
			if ( varNames != null && varObjects != null &&
				 varNames.length == varObjects.length )
			{
				for ( int i=0; i<varNames.length; i++ )
				{
					nameSpace.setVariable( varNames[ i ],
										   varObjects[ i ],
										   false );

					MainApp.info( "Providing script access for \"" +
								   varNames[ i ] +
								   "\"" );
				}
			}
		}
		catch ( Throwable e )
		{
			MainApp.error( e.getMessage() );
		}

		return nameSpace;
	}


	// +---------------------------------------------------------------------------+
	// |  getStdNameSpace                                                          |
	// +---------------------------------------------------------------------------+
	// |  Returns a namespace that has been initialized with a pre-defined set of  |
	// |  classes, packages and variable names.                                    |
	// |                                                                           |
	// |  @param sName - The name of the namespace.                                |
	// +---------------------------------------------------------------------------+
	public static bsh.NameSpace getStdNameSpace( String sName )
	{
		NameSpace nameSpace = new NameSpace( m_interpreter.getNameSpace(),
											 "NameSpace_" + sName );

		nameSpace.getClassManager().setClassLoader(
											ClassLoader.getSystemClassLoader() );

		try
		{
			nameSpace.getClassManager().addClassPath(
					( new java.io.File( MainApp.getAppPath() ) ).toURI().toURL() );

			nameSpace.getClassManager().addClassPath(
					( new java.io.File( MainApp.getScriptsPath() ) ).toURI().toURL() );

			if ( !m_externClassPaths.isEmpty() )
			{
				for ( URL classPath : m_externClassPaths )
				{
					nameSpace.getClassManager().addClassPath( classPath );
				}
			}

			setStdInit( nameSpace );
		}
		catch ( Throwable e )
		{
			MainApp.error( e.getMessage() );
		}

		return nameSpace;
	}

	// +---------------------------------------------------------------------------+
	// |  setStdInit                                                               |
	// +---------------------------------------------------------------------------+
	// |  Sets the specified namespace with a pre-defined standard set of classes, |
	// |  packages and variable names.                                             |
	// |                                                                           |
	// |  @param nameSpace - The namespace to be initialized.                      |
	// +---------------------------------------------------------------------------+
	public static void setStdInit( NameSpace nameSpace )
	{
		try
		{
			//  Add the scripts folder path along with all sub-folder names
			// +-----------------------------------------------------------+
			File fScriptPath = new File( MainApp.getScriptsPath() );

			nameSpace.importCommands( fScriptPath.getName() );

			File[] fScriptPathList = fScriptPath.listFiles();

			for ( int i=0; i<fScriptPathList.length; i++ )
			{
				if ( fScriptPathList[ i ].isDirectory() )
				{
					nameSpace.importCommands( fScriptPathList[ i ].getName() );
				}
			}


			//  Add variables
			// +-----------------------------------------------------------+
			nameSpace.setVariable( "logger",     MainApp.getLogger(),     false );
			nameSpace.setVariable( "interp",     m_interpreter,           false );
			nameSpace.setVariable( "bitmapPath", MainApp.getBitmapPath(), false );
			nameSpace.setVariable( "xmlPath",    MainApp.getXMLPath(),    false );

			nameSpace.setVariable( "ns", nameSpace, false );

			nameSpace.setVariable(
						"fitsHeaderFrame",
						MainApp.getMainFrame().getExposePanel().getFitsHeaderFrame(),
						false );


			//  Add default imports ( packages and classes )
			// +-----------------------------------------------------------+
			nameSpace.loadDefaultImports();


			//  Add packages
			// +-----------------------------------------------------------+
			nameSpace.importPackage( "java.awt.geom" );
			nameSpace.importPackage( "owl.plot" );


			//  Add classes
			// +-----------------------------------------------------------+
			nameSpace.importClass( "import arc.api.deinterlace.ArcDeinterlaceJAPI;" );
			nameSpace.importClass( "import arc.api.device.ArcDeviceJAPI;" );
			nameSpace.importClass( "import arc.api.device.ReplyException;" );
			nameSpace.importClass( "import arc.api.fits.ArcFitsFileJAPI;" );
			nameSpace.importClass( "import arc.api.tiff.ArcTiffFileJAPI;" );
			nameSpace.importClass( "import arc.api.image.ArcImageJAPI;" );

			nameSpace.importClass( "import owl.main.fits.FitsHeaderFrame;" );
			nameSpace.importClass( "import owl.main.MainApp;" );
		}
		catch ( Throwable e )
		{
			MainApp.error( e.getMessage() );
		}
	}

	// +---------------------------------------------------------------------------+
	// |  getNameSpaceInfo                                                         |
	// +---------------------------------------------------------------------------+
	// |  Returns a list of the specified namespace variables, methods and paths.  |
	// |  The first index ( 0 ) is the variable, method, etc name.                 |
	// |  The second index ( 1 ) is the variable, method, etc value.               |
	// |                                                                           |
	// |  @param nameSpace - The namespace to retrieve info for.                   |
	// +---------------------------------------------------------------------------+
	public static Object[][] getNameSpaceInfo( NameSpace nameSpace )
	{
		//  Get interpreter variable, method and classpath names
		// +-----------------------------------------------------+
		String[] varNames     = nameSpace.getVariableNames();
		String[] methodNames  = nameSpace.getMethodNames();
		URL[]    classPathURL = null;

		try
		{
			bsh.classpath.BshClassPath classPath =
								bsh.classpath.BshClassPath.getUserClassPath();

			classPathURL = classPath.getPathComponents();
		}
		catch ( bsh.ClassPathException cpe ) {}

		//  Create the 2-D object array
		// +-----------------------------------------------------+
		int dSize = ( varNames != null ? varNames.length : 0 ) +
					( methodNames != null ? methodNames.length : 0 ) +
					( classPathURL != null ? classPathURL.length : 0 ) + 1;

		Object[][] objInfo = new Object[ dSize ][ dSize ];

		int dIndex = 0;

		//  Add the namespace name
		// +-----------------------------------------------------+
		objInfo[ dIndex ][ 0 ] = "NameSpace Name";
		objInfo[ dIndex ][ 1 ] = nameSpace.getName();
		dIndex++;

		//  Add variable names if they exist
		// +-----------------------------------------------------+
		if ( varNames != null )
		{
			for ( int i=0; i<varNames.length; i++ )
			{
				try
				{
					String sVarVal = nameSpace.getVariable( varNames[ i ] ).toString();

					objInfo[ dIndex ][ 0 ] = "[ Variable ]  \"" + varNames[ i ] + "\"";
					objInfo[ dIndex ][ 1 ] = sVarVal;
					dIndex++;
				}
				catch ( bsh.UtilEvalError uee ) {}
			}
		}

		//  Add method names if they exist
		// +-----------------------------------------------------+
		if ( methodNames != null )
		{
			for ( int i=0; i<methodNames.length; i++ )
			{
				objInfo[ dIndex ][ 0 ] = "[ Method ]";
				objInfo[ dIndex ][ 1 ] = methodNames[ i ];
				dIndex++;
			}
		}

		//  Add classpath url's if they exist
		// +-----------------------------------------------------+
		if ( classPathURL != null )
		{
			for ( int i=0; i<classPathURL.length; i++ )
			{
				objInfo[ dIndex ][ 0 ] = "[ Class Path ]";
				objInfo[ dIndex ][ 1 ] = classPathURL[ i ].toString();
				dIndex++;
			}
		}

		return objInfo;
	}

	// +---------------------------------------------------------------------------+
	// |  getInterpreterInfo                                                       |
	// +---------------------------------------------------------------------------+
	// |  Returns a list of variables, methods and paths for the interpreter.      |
	// |  The first index ( 0 ) is the variable, method, etc name.                 |
	// |  The second index ( 1 ) is the variable, method, etc value.               |
	// +---------------------------------------------------------------------------+
	public static Object[][] getInterpreterInfo()
	{
		//  Get interpreter variable, method and classpath names
		// +-----------------------------------------------------+
		String[] varNames     = m_interpreter.getNameSpace().getVariableNames();
		String[] methodNames  = m_interpreter.getNameSpace().getMethodNames();
		URL[]    classPathURL = null;

		try
		{
			bsh.classpath.BshClassPath classPath =
								bsh.classpath.BshClassPath.getUserClassPath();

			classPathURL = classPath.getPathComponents();
		}
		catch ( bsh.ClassPathException cpe ) {}

		//  Create the 2-D object array
		// +-----------------------------------------------------+
		int dSize = ( varNames != null ? varNames.length : 0 ) +
					( methodNames != null ? methodNames.length : 0 ) +
					( classPathURL != null ? classPathURL.length : 0 ) + 1;

		Object[][] interpInfo = new Object[ dSize ][ dSize ];

		int dIndex = 0;

		//  Add the interpreter version
		// +-----------------------------------------------------+
		interpInfo[ dIndex ][ 0 ] = "Interpreter Version";
		interpInfo[ dIndex ][ 1 ] = Interpreter.VERSION;
		dIndex++;

		//  Add variable names if they exist
		// +-----------------------------------------------------+
		if ( varNames != null )
		{
			for ( int i=0; i<varNames.length; i++ )
			{
				try
				{
					String sVarVal = m_interpreter.getNameSpace().getVariable( varNames[ i ] ).toString();

					interpInfo[ dIndex ][ 0 ] = "[ Variable ]  \"" + varNames[ i ] + "\"";
					interpInfo[ dIndex ][ 1 ] = sVarVal;
					dIndex++;
				}
				catch ( bsh.UtilEvalError uee ) {}
			}
		}

		//  Add method names if they exist
		// +-----------------------------------------------------+
		if ( methodNames != null )
		{
			for ( int i=0; i<methodNames.length; i++ )
			{
				interpInfo[ dIndex ][ 0 ] = "[ Method ]";
				interpInfo[ dIndex ][ 1 ] = methodNames[ i ];
				dIndex++;
			}
		}

		//  Add classpath url's if they exist
		// +-----------------------------------------------------+
		if ( classPathURL != null )
		{
			for ( int i=0; i<classPathURL.length; i++ )
			{
				interpInfo[ dIndex ][ 0 ] = "[ Class Path ]";
				interpInfo[ dIndex ][ 1 ] = classPathURL[ i ].toString();
				dIndex++;
			}
		}

		return interpInfo;
	}

	// +---------------------------------------------------------------------------+
	// |  createNameSpaceName                                                      |
	// +---------------------------------------------------------------------------+
	// |  Returns a string that can be used as a namespace name.  The name has the |
	// |  following form:  "NameSpace_123456"                                      |
	// +---------------------------------------------------------------------------+
	private static String createNameSpaceName()
	{
		return ( "NameSpace_" + m_random.nextInt( 0xEFFFFFF ) );
	}

	// +---------------------------------------------------------------------------+
	// |  createNameSpaceName                                                      |
	// +---------------------------------------------------------------------------+
	// |  Returns a string that can be used as a namespace name.  The name has the |
	// |  following form:  "NameSpace_123456_sText"                                |
	// |                                                                           |
	// |  @param sText - Text to be appended to the generated namespace name.      |
	// +---------------------------------------------------------------------------+
	private static String createNameSpaceName( String sText )
	{
		return ( "NameSpace_" + m_random.nextInt( 0xEFFFFFF ) + "_" + sText );
	}
}
