package owl.main.setup;

import arc.api.device.ArcDeviceJAPI;
import arc.api.device.ArcReplyException;

import owl.gui.utils.OwlUtilities;
import owl.main.MainApp;


public class SetupRunnable implements Runnable
{
	private final int TDL_MAX = 0x1000000;

	private SetupFrame setupFrame;
	private boolean stop;

	public SetupRunnable( SetupFrame setupFrame )
	{
		this.setupFrame = setupFrame;
	}

	public void stop()
	{
		ArcDeviceJAPI.SetAbort( true );
		stop = true;
	}

	public void run()
	{
		boolean bTimFileLoaded = false;
		boolean bImageSizeSet  = false;
		boolean bPoweredOn     = false;

		Thread.currentThread().setName( "Owl - Setup Runnable" );

		setupFrame.setApplyAction( setupFrame.ABORT_ACTION );
		stop = false;

		try
		{
			if ( !ArcDeviceJAPI.IsOpen() )
			{
				throw new Exception( "Not connected to any device!" );
			}

			//
			// Reset the PCI(e) board. This prevents any previous settings
			// from affecting a fresh setup. For example, if 2x FO's has
			// been set, then a fresh setup will reset the controller back
			// to single FO, but not the PCI(e) board. This will fix that
			// by putting the PCI(e) back into single FO mode.
			//
			MainApp.infoStart( "Resetting PCI(e) board" );
			ArcDeviceJAPI.Reset();
			MainApp.infoEnd();

			if ( !stop && setupFrame.doResetController() )
			{
				MainApp.infoStart( "Resetting controller" );
				ArcDeviceJAPI.ResetController();
				MainApp.infoEnd();
			}

			if ( !stop && setupFrame.doHardwareTest() && setupFrame.doPCIHWTest() )
			{
				int dIncr = TDL_MAX / setupFrame.getPCIHWTCount();

				MainApp.infoStart( "Testing PCI(e) data link" );
				for ( int i=0; i<setupFrame.getPCIHWTCount(); i++ )
				{
					ArcDeviceJAPI.Cmd( ArcDeviceJAPI.PCI_ID, ArcDeviceJAPI.TDL, i * dIncr, i * dIncr );
					if ( stop ) { break; }
				}
				MainApp.infoEnd();
			}

			if ( !stop && setupFrame.doHardwareTest() && setupFrame.doTIMHWTest() )
			{
				int dIncr = TDL_MAX / setupFrame.getTIMHWTCount();

				MainApp.infoStart( "Testing TIM data link" );
				for ( int i=0; i<setupFrame.getTIMHWTCount(); i++ )
				{
					ArcDeviceJAPI.Cmd( ArcDeviceJAPI.TIM_ID, ArcDeviceJAPI.TDL, i * dIncr, i * dIncr );
					if ( stop ) { break; }
				}
				if ( !stop ) MainApp.infoEnd();
			}

			if ( !stop && setupFrame.doHardwareTest() && setupFrame.doUTLHWTest() )
			{
				int dIncr = TDL_MAX / setupFrame.getUTLHWTCount();

				MainApp.infoStart( "Testing UTIL data link" );
				for ( int i=0; i<setupFrame.getUTLHWTCount(); i++ )
				{
					ArcDeviceJAPI.Cmd( ArcDeviceJAPI.UTIL_ID, ArcDeviceJAPI.TDL, i * dIncr, i * dIncr );
					if ( stop ) { break; }
				}
				MainApp.infoEnd();
			}

			if ( !stop && setupFrame.doLoadTIM() )
			{
				MainApp.infoStart( "Loading TIM file" );

				if ( !OwlUtilities.grep( setupFrame.getTIMLoadFile(), "_START TIM" ) )
				{
					throw new Exception(
							"Invalid timing file .... does not start with \"_START TIM\"." );
				}

				ArcDeviceJAPI.LoadControllerFile( setupFrame.getTIMLoadFile(), 1 );
				bTimFileLoaded = true;

				MainApp.infoEnd();
			}

			if ( !stop && setupFrame.doLoadUTL() )
			{
				MainApp.infoStart( "Loading UTL file" );

				if ( !OwlUtilities.grep( setupFrame.getUTLLoadFile(), "_START UTIL" ) )
				{
					throw new Exception(
							"Invalid utility file .... does not start with \"_START UTIL\"." );
				}

				ArcDeviceJAPI.LoadControllerFile( setupFrame.getUTLLoadFile(), 1 );

				MainApp.infoEnd();
			}

			if ( !stop && setupFrame.doPowerOn() )
			{
				MainApp.infoStart( "Powering on controller" );
				ArcDeviceJAPI.Cmd( ArcDeviceJAPI.TIM_ID, ArcDeviceJAPI.PON, ArcDeviceJAPI.DON );
				MainApp.infoEnd();
				bPoweredOn = true;
			}

			if ( !stop && setupFrame.doImageSize() )
			{
				int dRows = setupFrame.getRowSize();
				int dCols = setupFrame.getColSize();

				bImageSizeSet = true;

				MainApp.infoStart( "Setting image dimensions" );
				ArcDeviceJAPI.SetImageSize( dRows, dCols );
				MainApp.infoEnd();
			}

			if ( !stop && ( bTimFileLoaded || bPoweredOn || bImageSizeSet ) )
			{
				SetupFrame.callSetupListeners( setupFrame );
			}

			MainApp.info( "Setup complete" );
		}
		catch ( ArcReplyException re )
		{
			MainApp.infoFail();
			MainApp.error( re.toString() );
		}
		catch ( Throwable e )
		{
			MainApp.infoFail();
			MainApp.error( "SetupThread error: " + e );
		}
		finally
		{
			ArcDeviceJAPI.SetAbort( false );
			setupFrame.setApplyAction( setupFrame.APPLY_ACTION );
		}
	}
}
